#include <msp/core/maputils.h>
#include "syntax.h"
#include "visitor.h"

using namespace std;

namespace Msp {
namespace GL {
namespace SL {

const Operator Operator::operators[] =
{
	{ "[", "]", 2, BINARY, LEFT_TO_RIGHT },
	{ "(", ")", 2, POSTFIX, LEFT_TO_RIGHT },
	{ ".", { }, 2, POSTFIX, LEFT_TO_RIGHT },
	{ "++", { }, 2, POSTFIX, LEFT_TO_RIGHT },
	{ "--", { }, 2, POSTFIX, LEFT_TO_RIGHT },
	{ "++", { }, 3, PREFIX, RIGHT_TO_LEFT },
	{ "--", { }, 3, PREFIX, RIGHT_TO_LEFT },
	{ "+", { }, 3, PREFIX, RIGHT_TO_LEFT },
	{ "-", { }, 3, PREFIX, RIGHT_TO_LEFT },
	{ "~", { }, 3, PREFIX, RIGHT_TO_LEFT },
	{ "!", { }, 3, PREFIX, RIGHT_TO_LEFT },
	{ "*", { }, 4, BINARY, ASSOCIATIVE },
	{ "/", { }, 4, BINARY, LEFT_TO_RIGHT },
	{ "%", { }, 4, BINARY, LEFT_TO_RIGHT },
	{ "+", { }, 5, BINARY, ASSOCIATIVE },
	{ "-", { }, 5, BINARY, LEFT_TO_RIGHT },
	{ "<<", { }, 6, BINARY, LEFT_TO_RIGHT },
	{ ">>", { }, 6, BINARY, LEFT_TO_RIGHT },
	{ "<", { }, 7, BINARY, LEFT_TO_RIGHT },
	{ ">", { }, 7, BINARY, LEFT_TO_RIGHT },
	{ "<=", { }, 7, BINARY, LEFT_TO_RIGHT },
	{ ">=", { }, 7, BINARY, LEFT_TO_RIGHT },
	{ "==", { }, 8, BINARY, LEFT_TO_RIGHT },
	{ "!=", { }, 8, BINARY, LEFT_TO_RIGHT },
	{ "&", { }, 9, BINARY, ASSOCIATIVE },
	{ "^", { }, 10, BINARY, ASSOCIATIVE },
	{ "|", { }, 11, BINARY, ASSOCIATIVE },
	{ "&&", { }, 12, BINARY, ASSOCIATIVE },
	{ "^^", { }, 13, BINARY, ASSOCIATIVE },
	{ "||", { }, 14, BINARY, ASSOCIATIVE },
	{ "?", ":", 15, TERNARY, RIGHT_TO_LEFT },
	{ "=", { }, 16, BINARY, RIGHT_TO_LEFT },
	{ "+=", { }, 16, BINARY, RIGHT_TO_LEFT },
	{ "-=", { }, 16, BINARY, RIGHT_TO_LEFT },
	{ "*=", { }, 16, BINARY, RIGHT_TO_LEFT },
	{ "/=", { }, 16, BINARY, RIGHT_TO_LEFT },
	{ "%=", { }, 16, BINARY, RIGHT_TO_LEFT },
	{ "<<=", { }, 16, BINARY, RIGHT_TO_LEFT },
	{ ">>=", { }, 16, BINARY, RIGHT_TO_LEFT },
	{ "&=", { }, 16, BINARY, RIGHT_TO_LEFT },
	{ "^=", { }, 16, BINARY, RIGHT_TO_LEFT },
	{ "|=", { }, 16, BINARY, RIGHT_TO_LEFT },
	{ ",", { }, 17, BINARY, LEFT_TO_RIGHT },
	{ { 0 }, { }, 18, NO_OPERATOR, LEFT_TO_RIGHT }
};

const Operator &Operator::get_operator(const string &token, Type type)
{
	for(const Operator *i=operators; i->type; ++i)
		if(i->type==type && i->token==token)
			return *i;
	throw key_error(token);
}


template<typename C>
NodeContainer<C>::NodeContainer(const NodeContainer &c):
	C(c)
{
	for(typename C::iterator i=this->begin(); i!=this->end(); ++i)
		*i = (*i)->clone();
}


Block::Block():
	use_braces(false),
	parent(0)
{ }

Block::Block(const Block &other):
	Node(other),
	body(other.body),
	use_braces(other.use_braces),
	parent(0)
{ }

void Block::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


Expression::Expression():
	oper(0),
	type(0),
	lvalue(false)
{ }


void Literal::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


VariableReference::VariableReference():
	declaration(0)
{ }

VariableReference::VariableReference(const VariableReference &other):
	Expression(other),
	name(other.name),
	declaration(0)
{ }

void VariableReference::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


InterfaceBlockReference::InterfaceBlockReference():
	declaration(0)
{ }

InterfaceBlockReference::InterfaceBlockReference(const InterfaceBlockReference &other):
	Expression(other),
	name(other.name),
	declaration(0)
{ }

void InterfaceBlockReference::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


MemberAccess::MemberAccess():
	declaration(0),
	index(-1)
{ }

MemberAccess::MemberAccess(const MemberAccess &other):
	Expression(other),
	left(other.left),
	member(other.member),
	declaration(0),
	index(-1)
{ }

void MemberAccess::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


Swizzle::Swizzle():
	count(0)
{
	fill(components, components+4, 0);
}

void Swizzle::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


void UnaryExpression::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


void BinaryExpression::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


Assignment::Assignment():
	self_referencing(false)
{ }

Assignment::Assignment(const Assignment &other):
	BinaryExpression(other),
	self_referencing(other.self_referencing)
{ }

void Assignment::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


Assignment::Target::Target(Statement *d):
	declaration(d),
	chain_len(0)
{
	fill(chain, chain+7, 0);
}

bool Assignment::Target::operator<(const Target &other) const
{
	if(declaration!=other.declaration)
		return declaration<other.declaration;
	for(unsigned i=0; (i<7 && i<chain_len && i<other.chain_len); ++i)
		if(chain[i]!=other.chain[i])
			return chain[i]<other.chain[i];
	return chain_len<other.chain_len;
}


void TernaryExpression::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


FunctionCall::FunctionCall():
	constructor(false),
	declaration(0)
{ }

FunctionCall::FunctionCall(const FunctionCall &other):
	Expression(other),
	name(other.name),
	constructor(other.constructor),
	arguments(other.arguments),
	declaration(0)
{ }

void FunctionCall::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


void ExpressionStatement::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


void Import::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


void Precision::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


void Layout::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


void InterfaceLayout::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


BasicTypeDeclaration::BasicTypeDeclaration():
	kind(ALIAS),
	size(0),
	base_type(0)
{ }

BasicTypeDeclaration::BasicTypeDeclaration(const BasicTypeDeclaration &other):
	TypeDeclaration(other),
	kind(other.kind),
	size(other.size),
	base(other.base),
	base_type(0)
{ }

void BasicTypeDeclaration::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


ImageTypeDeclaration::ImageTypeDeclaration():
	dimensions(TWO),
	array(false),
	sampled(true),
	shadow(false)
{ }

void ImageTypeDeclaration::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


StructDeclaration::StructDeclaration():
	interface_block(0)
{
	members.use_braces = true;
}

StructDeclaration::StructDeclaration(const StructDeclaration &other):
	TypeDeclaration(other),
	members(other.members),
	interface_block(0)
{ }

StructDeclaration::~StructDeclaration()
{
	if(interface_block && interface_block->struct_declaration==this)
		interface_block->struct_declaration = 0;
}

void StructDeclaration::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


VariableDeclaration::VariableDeclaration():
	constant(false),
	array(false),
	type_declaration(0),
	linked_declaration(0)
{ }

VariableDeclaration::VariableDeclaration(const VariableDeclaration &other):
	Statement(other),
	layout(other.layout),
	constant(other.constant),
	sampling(other.sampling),
	interpolation(other.interpolation),
	interface(other.interface),
	precision(other.precision),
	type(other.type),
	name(other.name),
	array(other.array),
	array_size(other.array_size),
	init_expression(other.init_expression),
	type_declaration(0),
	linked_declaration(0)
{ }

VariableDeclaration::~VariableDeclaration()
{
	if(linked_declaration && linked_declaration->linked_declaration==this)
		linked_declaration->linked_declaration = 0;
}

void VariableDeclaration::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


InterfaceBlock::InterfaceBlock():
	array(false),
	type_declaration(0),
	struct_declaration(0),
	linked_block(0)
{ }

InterfaceBlock::InterfaceBlock(const InterfaceBlock &other):
	Statement(other),
	interface(other.interface),
	block_name(other.block_name),
	members(other.members),
	instance_name(other.instance_name),
	array(other.array),
	type_declaration(0),
	struct_declaration(0),
	linked_block(0)
{ }

InterfaceBlock::~InterfaceBlock()
{
	if(linked_block && linked_block->linked_block==this)
		linked_block->linked_block = 0;
	if(struct_declaration && struct_declaration->interface_block==this)
		struct_declaration->interface_block = 0;
}

void InterfaceBlock::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


FunctionDeclaration::FunctionDeclaration():
	virtua(false),
	overrd(false),
	definition(0),
	return_type_declaration(0)
{ }

FunctionDeclaration::FunctionDeclaration(const FunctionDeclaration &other):
	Statement(other),
	return_type(other.return_type),
	name(other.name),
	parameters(other.parameters),
	virtua(other.virtua),
	overrd(other.overrd),
	body(other.body),
	signature(other.signature),
	definition(other.definition==&other ? this : 0),
	return_type_declaration(0)
{ }

void FunctionDeclaration::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


void Conditional::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


void Iteration::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


void Passthrough::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


void Return::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


void Jump::visit(NodeVisitor &visitor)
{
	visitor.visit(*this);
}


Stage::Stage(Stage::Type t):
	type(t),
	previous(0)
{ }

const char *Stage::get_stage_name(Type type)
{
	static const char *names[] = { "shared", "vertex", "geometry", "fragment" };
	return names[type];
}


Module::Module():
	shared(Stage::SHARED)
{ }


string get_unused_variable_name(const Block &block, const string &base)
{
	string name = base;

	unsigned number = 1;
	unsigned base_size = name.size();
	while(1)
	{
		bool unused = true;
		for(const Block *b=&block; (unused && b); b=b->parent)
			unused = !b->variables.count(name);
		if(unused)
			return name;

		name.erase(base_size);
		name += format("_%d", number);
		++number;
	}
}

bool is_same_type(const TypeDeclaration &type1, const TypeDeclaration &type2)
{
	if(const BasicTypeDeclaration *basic1 = dynamic_cast<const BasicTypeDeclaration *>(&type1))
	{
		const BasicTypeDeclaration *basic2 = dynamic_cast<const BasicTypeDeclaration *>(&type2);
		if(!basic2)
			return false;

		if(basic1->kind!=basic2->kind || basic1->size!=basic2->size)
			return false;

		if(basic1->base_type && basic2->base_type)
			return is_same_type(*basic1->base_type, *basic2->base_type);
		else
			return (!basic1->base_type && !basic2->base_type);
	}
	else if(const ImageTypeDeclaration *image1 = dynamic_cast<const ImageTypeDeclaration *>(&type1))
	{
		const ImageTypeDeclaration *image2 = dynamic_cast<const ImageTypeDeclaration *>(&type2);
		if(!image2)
			return false;

		if(image1->dimensions!=image2->dimensions || image1->array!=image2->array)
			return false;
		if(image1->sampled!=image2->sampled || image1->shadow!=image2->shadow)
			return false;

		if(image1->base_type && image2->base_type)
			return is_same_type(*image1->base_type, *image2->base_type);
		else
			return (!image1->base_type && !image2->base_type);
	}
	else if(const StructDeclaration *strct1 = dynamic_cast<const StructDeclaration *>(&type1))
	{
		const StructDeclaration *strct2 = dynamic_cast<const StructDeclaration *>(&type2);
		if(!strct2)
			return false;

		NodeList<Statement>::const_iterator i = strct1->members.body.begin();
		NodeList<Statement>::const_iterator j = strct2->members.body.begin();
		for(; (i!=strct1->members.body.end() && j!=strct2->members.body.end()); ++i, ++j)
		{
			const VariableDeclaration *var1 = dynamic_cast<const VariableDeclaration *>(i->get());
			const VariableDeclaration *var2 = dynamic_cast<const VariableDeclaration *>(j->get());
			if(!var1 || !var1->type_declaration || !var2 || !var2->type_declaration)
				return false;
			if(!is_same_type(*var1->type_declaration, *var2->type_declaration))
				return false;
			if(var1->name!=var2->name || var1->array!=var2->array)
				return false;
			// TODO Compare array sizes
			// TODO Compare layout qualifiers for interface block members
		}

		return (i==strct1->members.body.end() && j==strct2->members.body.end());
	}
	else
		return false;
}

int get_layout_value(const Layout &layout, const string &name, int def_value)
{
	for(vector<Layout::Qualifier>::const_iterator i=layout.qualifiers.begin(); i!=layout.qualifiers.end(); ++i)
		if(i->name==name)
			return i->value;
	return def_value;
}

void add_to_chain(Assignment::Target &target, Assignment::Target::ChainType type, unsigned index)
{
	if(target.chain_len<7)
		target.chain[target.chain_len] = type | min<unsigned>(index, 0x3F);
	++target.chain_len;
}

} // namespace SL
} // namespace GL
} // namespace Msp
